/*--------------------------------------------------------------------
  Terminal Module.
    based on NxEdit2.04    ::term.c
             FD clone 1.03g::term.c

    Copyright (C) 1999
        SASAKI Shunsuke <ele@pop17.odn.ne.jp> All rights reserved. 
--------------------------------------------------------------------*/
#include	"config.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	<signal.h>
#include	<sys/types.h>
#include	<errno.h>


#ifdef	HAVE_STDARG_H
#	include	<stdarg.h>
#else
#	ifdef	HAVE_VARARGS_H
#		include	<varargs.h>
#	endif
#endif

#ifdef	HAVE_FCNTL_H
#	include	<fcntl.h>
#endif

#if	defined(HAVE_SELECT) &&	defined(HAVE_SYS_TIME_H)
#	include <sys/time.h>
#endif


#ifdef	HAVE_TERMCAP_H
#	include	<termcap.h>
#endif

#ifdef	HAVE_TERMIOS_H
#	include	<termios.h>
#	ifdef	HAVE_TERMIO_H
#		undef	HAVE_TERMIO_H
#	endif
#else
#	include	<termio.h>
#endif


/*
	#ifdef	NOERRNO
	extern int errno;
	#endif
*/



#ifndef	TRUE
#define	TRUE	!FALSE
#endif
#ifndef	FALSE
#define	FALSE	0
#endif

typedef	int				bool;


#include "term.h"


#define	LN_dispbuf	2048
#define	TERMCAPSIZE	2048

#define	STDIN		0
#define	STDOUT		1
#define	STDERR		2
#define	TTYNAME		"/dev/tty"

#ifndef	IEXTEN
#define	IEXTEN		0
#endif


static	void	term_getent();

#define	term_tput(str)	tputs(str, 1, term_dputc)
#define	term_tputs(str)	tputs(str, term.sizey, term_dputc)

/*
#ifdef __linux__
speed_t	ospeed= 0;
#else
short	ospeed= 0;
#endif
*/

char	PC= '\0';
char	*BC= NULL;
char	*UP= NULL;

#if	!defined (TIOCSTI)
static u_char ungetbuf[10];
static int ungetnum = 0;
#endif




typedef	struct
{
#ifdef	HAVE_TERMIOS_H
	struct termios	tty;
#else
	struct termio	tty;
#endif

	int 	ttyio;
	FILE	*ttyfp;
	char	*name;
	int 	mode;

	int 	sizey;		// !!??
	int 	sizex;		// !!??

	char	tbuf[TERMCAPSIZE+1];

	char	*t_init;
	char	*t_keypad;
	char	*t_nokeypad;
	char	*t_end;
	char	*c_locate;

	char	*t_clear;
	char	*t_normal;
	char	*t_reverse;
	char	*t_underline;
	char	*t_bold;
	char	*l_insert;
	char	*l_delete;
	char	*l_clear;
	char	*end_underline;

	char	*t_highcursor;
	char	*t_nocursor;
	char	*t_normalcursor;

	char	*t_vbell;
}	term_t;

term_t	term;


typedef	struct
{
	int 	x,y;	/* latest cursor location */
	color_t	cl;		/* latest color */
}	termm_t;
termm_t	termm;

static	int 	exit_failed=FALSE;


int 	GetMinRow()		{return 1;}
int 	GetMinCol()		{return 0;}

int 	GetMaxRow()		{return term.sizey-1;}
int 	GetRowWidth()	{return term.sizey-1;}
int 	GetMaxCol()		{return term.sizex-1;}
int 	GetColWidth()	{return term.sizex-1;}




void	term_getwsize()
{
	int x, y;

	x = y = -1;

#ifdef	TIOCGWINSZ
		{
		 struct winsize ws;

		 if (ioctl(term.ttyio, TIOCGWINSZ, &ws)>=0)
		 	{
		 	 x = ws.ws_col;
		 	 y = ws.ws_row;
		 	}
		}
#endif
#ifdef	WIOCGETD
	if (x<0&& y<0)
		{
		 struct uwdate wd;

		 if (ioctl(term.ttyio, WIOCGETD, &ws)>=0)
		 	{
		 	 x = ws.uw_width / ws.uw_hs;
		 	 y = ws.uw_height / ws.uw_vs;
		 	}
		}
#endif

	if (x>0 && y>0)
		{
		 term.sizex = x;
		 term.sizey = y;
		}
}

void	term_setmode(int mode)
{
	if ((term.mode& TM_keypad)&& !(mode& TM_keypad))
		term_tputs(term.t_nokeypad);

	if ((mode& TM_keypad)&& !(term.mode& TM_keypad))
		term_tputs(term.t_keypad);


	if ((term.mode& TM_init)&& !(mode& TM_init))
		{
		 term_color_normal();
		 term_cls();
		 if (term.mode& TM_tinit)
		 	term_tputs(term.t_end);
		}

	if ((mode& TM_init)&& !(term.mode& TM_init))
		{
		 termm.x=-1;
		 termm.y=-1;
		 termm.cl.fc=AC_white;
		 termm.cl.bc=AC_black;
		 termm.cl.rf=FALSE;
		 termm.cl.uf=FALSE;
		 termm.cl.bf=FALSE;

		 if (mode& TM_tinit)
		 	term_tputs(term.t_init);
		 term_tputs(term.t_normal);
		 term_csrn();

		 term_getwsize();
		}

	term_flush();
	term.mode=mode;
}

int 	term_getmode()
{
	return term.mode;
}




void	term_stop()	/* atexitΤŪ˸ƤɬפϤʤ */
{
	if (exit_failed)
		return;
	exit_failed=TRUE;


	term_setmode(0);

#ifdef	HAVE_TERMIOS_H
	if (tcsetattr(term.ttyio, TCSAFLUSH, &term.tty)<0)
		error("term_fin():tcsetattr");
#else
	if (ioctl(term.ttyio, TCSETAF, &term.tty)<0)
		error("term_fin():TCSETAF");
#endif
}

void	term_start(int mode)
{
	struct termios	tty;

	memcpy(&tty, &term.tty, sizeof(struct termios));
	tty.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL | ISIG | ICANON | IEXTEN);
	tty.c_iflag |= IGNBRK;
	tty.c_iflag &= ~(ICRNL | BRKINT | IXON);
	tty.c_oflag &= ~(ONLCR| OPOST);
	tty.c_cc[VTIME]= 0;
	tty.c_cc[VMIN]= 1;

#ifdef	HAVE_TERMIOS_H
//	ospeed = cfgetospeed(&term.tty);
	if (tcsetattr(term.ttyio, TCSAFLUSH, &tty)<0)
		error("term_init():tcsetattr");
#else
//	ospeed = term.tty.c_cflag & CBAUD;
	if (ioctl(term.ttyio, TCGETA, &tty)<0)
		error("term_fin():TCSETAF");
#endif

	term_setmode(mode);
	exit_failed=FALSE;
	atexit(term_stop);
}

void	term_init()
{
	term.mode=0;

	term.ttyio=open(TTYNAME, O_RDWR, 0600);
	if (term.ttyio<0)
		term.ttyio=STDERR;
	term.ttyfp=fdopen(term.ttyio, "w+");
	if (term.ttyfp==NULL)
		error("term_init():fdopen");

	if (tcgetattr(term.ttyio, &term.tty)<0)
		error("term_init():tcgetattr");

	term.name= getenv("TERM");
	if (term.name==NULL)
		term.name="vt100";

	term_getent();
}

void	term_kflush()
{
	tcflush(term.ttyio, TCIFLUSH);
}

int 	term_dputc(int c)
{
	return(fputc(c, term.ttyfp));
}

int 	term_getch()
{
	u_char ch;
	int i;

	while ((i=read(term.ttyio, &ch, sizeof(u_char))) < 0 && errno == EINTR);
	if (i < sizeof(u_char)) return(EOF);
	return((int)ch);
}

int 	term_kbhit(u_long usec)
{
	fd_set readfds;
	struct timeval timeout;

	timeout.tv_sec = 0;
	timeout.tv_usec = usec;
	FD_ZERO(&readfds);
	FD_SET(STDIN, &readfds);

	return(select(1, &readfds, NULL, NULL, &timeout));
}

void	term_flush()	{fflush(term.ttyfp);}
void	term_insln()	{term_tput(term.l_insert);}
void	term_delln()	{term_tput(term.l_delete);}
void	term_bell()		{term_tput(term.t_vbell);}
void	term_csrf()		{term_tput(term.t_highcursor);}
void	term_csrn()		{term_tput(term.t_normalcursor);}
void	term_csrh()		{term_tput(term.t_nocursor);}





char	*term_setent(const char *str,const char *dstr)
{
	char	strbuf[TERMCAPSIZE+1];
	char	*p, *cp, *term;

	p=strbuf;
	cp=tgetstr(str, &p);

	if (cp==NULL)
		return (char *)dstr;

	term=(char *)malloc(strlen(cp)+1);
	if (term==NULL)
		error(NULL);
	strcpy(term, cp);
	return term;
}



void	term_getent()
{
	char	*p;
	int 	i;

	if (tgetent(term.tbuf, term.name) <= 0)
		{
//		 warning("No termcap presented.");
		 term.sizey=24;
		 term.sizex=79;
		} else
		{
		 term.sizey = tgetnum("li");
		 term.sizex = tgetnum("co");
		 if (!tgetflag("am"))
		 	term.sizex--;

		 PC=*term_setent("pc", ""); //!! ݤʤΤFreeƤʤ
		 BC=term_setent("bc", "\010");
		 UP=term_setent("up", "\033[A");

		 term.t_init         =term_setent("ti","");
		 term.t_end          =term_setent("te","");
		 term.t_keypad       =term_setent("ks","\033[?1h\033=");
		 term.t_nokeypad     =term_setent("ke","\033[?1l\033>");
		 term.t_normalcursor =term_setent("ve","\033[?25h");
		 term.t_highcursor   =term_setent("vs","\033[?25h");
		 term.t_nocursor     =term_setent("vi","\033[?25l");
		 term.t_vbell        =term_setent("vb","\007");
		 term.t_clear        =term_setent("cl","\033[;H\033[2J");
		 term.t_normal       =term_setent("me","\033[m");
		 term.t_bold         =term_setent("md","\033[1m");
		 term.t_reverse      =term_setent("mr","\033[7m");
		 term.t_underline    =term_setent("us","\033[4m");
		 term.end_underline  =term_setent("ue","\033[m");
		 term.l_clear        =term_setent("ce","\033[K");
		 term.l_insert       =term_setent("al","\033[L");
		 term.l_delete       =term_setent("dl","\033[M");
		 term.c_locate       =term_setent("cm","\033[%i%d;%dH");
		}
}

int 	term_ungetch(c)
u_char c;
{
#ifdef	TIOCSTI
	ioctl(term.ttyio, TIOCSTI, &c);
#else
	if (ungetnum >= sizeof(ungetbuf) / sizeof(u_char) - 1) return(EOF);
	ungetbuf[ungetnum++] = c;
#endif
	return(c);
}

#define	SENSEPERSEC	50
#define	WAITKEYPAD	360		/* msec */











void	term_locate(int y,int x)
{
	if (termm.x==x&& termm.y==y)
		return;

	if (x==0&& termm.y+1==y)
		term_tputs("\r\n"); else
		term_tputs(tgoto(term.c_locate, x, y));

	termm.x=x;
	termm.y=y;
}

void	term_cls()
{
	term_color_normal();
	term_tput(term.t_clear);
}

void	term_clrtoe()
{
	term_color_normal();
	term_tput(term.l_clear);
}

void	term_movex(int x)
{
	termm.x+=x;
}

int 	term_puts(char *s)
{
	term_movex(kanji_poscanon(strlen(s),s));
	return(fputs(s, term.ttyfp));
}

int 	term_putch(int c)
{
	term_movex(1);
	return term_dputc(c);
}

int 	term_printf(const char *fmt, ...)
{
	va_list	args;
	char	buf[LN_dispbuf+1];

	va_start(args,fmt);
	vsprintf(buf, fmt, args);
	va_end(args);

	term_puts(buf);
	return(strlen(buf));
}



void	col_normal(color_t *clp)
{
	memset(clp, 0, sizeof(color_t));
	clp->fc=AC_white;
	clp->bc=AC_black;
	clp->rf=FALSE;
	clp->uf=FALSE;
	clp->bf=FALSE;
}

bool	col_cmp(color_t cola,color_t colb)
{
	return cola.fc==colb.fc&& cola.bc==colb.bc&& cola.rf==colb.rf
		&& cola.uf==colb.uf&& cola.bf==colb.bf;
}


void	term_color(color_t cl)
{
	char	buf[20+1];

	if (termm.cl.uf&&!cl.uf)
		{
		 term_tput(term.end_underline);
		 if (strcmp(term.t_normal, term.end_underline)==0)
		 	col_normal(&termm.cl);
		}

	if ((termm.cl.rf&&!cl.rf)||(termm.cl.bf&&!cl.bf))
		{
		 term_tput(term.t_normal);
		 col_normal(&termm.cl);
		}


	if (term.mode& TM_ansicolor)
		{
		 bool	f;

		 strcpy(buf,"\x1b[");

		 f=FALSE;
		 if (termm.cl.fc!=cl.fc)
		 	{
		 	 sprintf(buf+strlen(buf), "3%d", cl.fc);
		 	 termm.cl.fc=cl.fc;
		 	 f=TRUE;
		 	 }

		 if (termm.cl.bc!=cl.bc)
		 	{
		 	 if (f)
		 	 	strcat(buf, ";");
		 	 sprintf(buf+strlen(buf),"4%d",cl.bc);
		 	 termm.cl.bc=cl.bc;
		 	 f=TRUE;
		 	}

		 if (f)
		 	{
		 	 strcat(buf,"m");
		 	 term_tput(buf);
		 	}
		}



	if (!termm.cl.rf&&cl.rf)
		term_tput(term.t_reverse);
	if (!termm.cl.uf&&cl.uf)
		term_tput(term.t_underline);
	if (!termm.cl.bf&&cl.bf)
		term_tput(term.t_bold);

	termm.cl.rf=cl.rf;
	termm.cl.uf=cl.uf;
	termm.cl.bf=cl.bf;

}

void	term_color_reverse()
{
	color_t	cl;

	cl=termm.cl;
	cl.rf=TRUE;

	term_color(cl);
}

void	term_color_underline()
{
	color_t	cl;

	cl=termm.cl;
	cl.uf=TRUE;

	term_color(cl);
}

void	term_color_bold()
{
	color_t	cl;

	cl=termm.cl;
	cl.bf=TRUE;

	term_color(cl);
}

void	term_color_fc(int fc)
{
	color_t	cl;

	cl=termm.cl;
	cl.fc=fc;

	term_color(cl);
}

void	term_color_normal()
{
	color_t	cl;

	col_normal(&cl);
	term_color(cl);
}

color_t	term_cftocol(const char *s)
{
	color_t	cl;
	bool	f;

	col_normal(&cl);

	if (s==NULL)
		return cl;

	f=FALSE;
	for (;*s!='\0';++s)
		{
		 switch(toupper(*s))
		 	{
		  case 'U':
		  	 cl.uf=TRUE;
		  	 continue;
		  case 'R':
		  	 cl.rf=TRUE;
		  	 continue;
		  case 'B':
		  	 cl.bf=TRUE;
		  	 continue;
		 	}

		 if (f)
		 	cl.bc=*s-'0'; else
		 	{
		 	 cl.fc=*s-'0';
		 	 f=TRUE;
		 	}
		}
	return cl;
}





















#define	MAX_esc	1024
typedef	struct
{
	char	*seq;
	int 	scode;

	int 	len;
}	esc_t;

static	esc_t	esc[MAX_esc];
static	int 	esc_num=0;


void	term_escset(int n, const char *e, const char *d)
{
	const	char	*p;
	char	*tp,buf[TERMCAPSIZE+1];

	if (esc_num>=MAX_esc || *e=='\0')
		return;

	esc[esc_num].scode= n;
	switch (*e)
		{
	 case '&':
		 tp=buf;
		 p=tgetstr(e+1, &tp);
		 if (p==NULL|| *p=='\0')
		 	{
		 	 if (d==NULL || *d=='\0')
		 	 	return;
		 	 p=d;
		 	}
		 break;

	 default:
	 	 p=e;
	 	}

	esc[esc_num].seq=(char *)malloc(strlen(p)+1);	// !!memory chk
	strcpy(esc[esc_num].seq, p);

	esc[esc_num].len  = strlen(esc[esc_num].seq);

//fprintf(stderr, "%s\n", esc[esc_num].seq );

	++esc_num;
}

int 	term_escdefault_cmp(const void *v,const void *w)
{
	esc_t	*kp_v,*kp_w;

	return strcmp(((esc_t *)v)->seq, ((esc_t *)w)->seq);
}

void	term_escdefault()
{
	if (esc_num==0)
		{
		 char	*p, buf[1024+1];

		 keys_set("[UP]",        "&ku", "\033OA");
		 keys_set("[DOWN]",      "&kd", "\033OB");
		 keys_set("[RIGHT]",     "&kr", "\033OC");
		 keys_set("[LEFT]",      "&kl", "\033OD");
		 keys_set("[HOME]",      "&kh", "\033[4~");
		 keys_set("[BS]",        "&kb", "\010");
		 keys_set("[DEL]",       "&kD", "\177");
		 keys_set("[INS]",       "&kI", "\033[2~");
		 keys_set("[Page_Up]",   "&kP", "\033[5~");
		 keys_set("[Page_Down]", "&kN", "\033[6~");
		 keys_set("[END]",       "&@7", "\033[1~");
//		 keys_set("[CR]",        "&@8","\033[9~");

		 keys_set("[F01]",       "&k1", "\033[11~");
		 keys_set("[F02]",       "&k2", "\033[12~");
		 keys_set("[F03]",       "&k3", "\033[13~");
		 keys_set("[F04]",       "&k4", "\033[14~");
		 keys_set("[F05]",       "&k5", "\033[15~");
		 keys_set("[F06]",       "&k6", "\033[17~");
		 keys_set("[F07]",       "&k7", "\033[18~");
		 keys_set("[F08]",       "&k8", "\033[19~");
		 keys_set("[F09]",       "&k9", "\033[20~");
		 keys_set("[F10]",       "&k0", "\033[21~");
		 keys_set("[F11]",       "&ka", "\033[23~");
		 keys_set("[F12]",       "&kb", "\033[24~");

		 keys_set("[F01]",       "&l1", "\033OP");
		 keys_set("[F02]",       "&l2", "\033OQ");
		 keys_set("[F03]",       "&l3", "\033OR");
		 keys_set("[F04]",       "&l4", "\033OS");
		 keys_set("[F05]",       "&l5", "\033OT");
		 keys_set("[F06]",       "&l6", "\033OU");
		 keys_set("[F07]",       "&l7", "\033OV");
		 keys_set("[F08]",       "&l8", "\033OW");
		 keys_set("[F09]",       "&l9", "\033OX");
		 keys_set("[F10]",       "&l0", "\033OY");

		 keys_set("*",           "$?(1b)Oj", "");
		 keys_set("+",           "$?(1b)Ok", "");
		 keys_set(",",           "$?(1b)Ol", "");
		 keys_set("-",           "$?(1b)Om", "");
		 keys_set(".",           "$?(1b)On", "");
		 keys_set("/",           "$?(1b)Oo", "");
		 keys_set("0",           "$?(1b)Op", "");
		 keys_set("1",           "$?(1b)Oq", "");
		 keys_set("2",           "$?(1b)Or", "");
		 keys_set("3",           "$?(1b)Os", "");
		 keys_set("4",           "$?(1b)Ot", "");
		 keys_set("5",           "$?(1b)Ou", "");
		 keys_set("6",           "$?(1b)Ov", "");
		 keys_set("7",           "$?(1b)Ow", "");
		 keys_set("8",           "$?(1b)Ox", "");
		 keys_set("9",           "$?(1b)Oy", "");
//		 keys_set("[CR]",        "$?(1b)OM", "");


		 p=getenv("NE_TERM");
		 if (p!=NULL)
		 	sprintf(buf, "term.%s", p); else
		 	{
		 	 p=getenv("TERM");
		 	 if (p!=NULL)
		 	 	sprintf(buf, "term.%s", p);
		 	}
		 if (p!=NULL)
		 	config_read(buf);

		}


	qsort(esc, esc_num, sizeof(esc_t), term_escdefault_cmp);
}


int 	term_inkey()
{
	int i, j, ch, key, start, end, s, e;

	term_flush();

	for (;;)
		{
#ifndef	TIOCSTI
		 if (ungetnum > 0) 
		 	{
		 	 ch = (int)ungetbuf[--ungetnum];
		 	 return(ch);
		 	}
#endif
		 key = term_kbhit(1000000L/SENSEPERSEC);
		 if (key==-1&& errno==EBADF)
		 	error("");
		 if (key!=0)
		 	break;
		}


	key = ch = term_getch();
	start = 0;
	end   = esc_num-1;

	for (i = 0; ; i++) 
	{
		s = start;
		e = end;
		start = end = -1;
		for (j = s; j <= e; j++) 
		{
			if (esc[j].len > i && ch == esc[j].seq[i]) 
			{
				if (esc[j].len == i + 1) return(esc[j].scode);
				end = j;
			}
			else if (end < 0) start = j + 1;
			else break;
		}
		if (start < 0) start = s;
		if (start > end || !term_kbhit(WAITKEYPAD * 1000L))
		{
		 	if (i==0) return(key);
			term_ungetch(ch);
			--i;
			for (;i>0;--i) term_ungetch(esc[j-1].seq[i]);
			return (esc[j-1].seq[0]);
		}
		ch = term_getch();
	}
}



/* !! ¸:  kanji_poscanon, keys_set, error */
